#include "pTPM.h"
#include "tss2/tss2_common.h"
#include "tss2/tss2_tpm2_types.h"
#include "tss2/tss2-tcti-tabrmd.h"
#include "stdio.h"
#include "stdlib.h"
#include <inttypes.h>
#include <string.h>
#include <errno.h>
#include <iostream>
#include <unordered_map>
#include <iterator>
#include "MapKeyClass.h"
#include "math.h"
#include "tss2/tss2_mu.h"


using namespace std;

// Constructor overload
pTPM::pTPM(int number)
{
    // Initialization of the contexts
    Init_Tcti_Tabrmd_Context();
    Init_Esys_Context();

    fprintf(stdout,"TPM instantiation successfull \n");

    identifier=number;

    // Initialization of the index
    freeIndex = TPM2_NV_INDEX_FIRST;
}

// Destructore overload
pTPM::~pTPM()
{
    // Finalization of the contexts
    Finalize_Esys_Context();
    Finalize_Tcti_Tabrmd_Context();

    fprintf(stdout,"TPM finalization successfull \n");
}

/*/////////////////////////////////////////////////////////
                    PRIVATE METHODS
/////////////////////////////////////////////////////////*/

void pTPM::Init_Tcti_Tabrmd_Context()
{
    TSS2_RC rc;
    size_t context_size;

    //Define allocation for the mssim TCTI
    /* 
        TSS2_RC Tss2_Tcti_Tabrmd_Init (TSS2_TCTI_CONTEXT *tcti_context, size_t *size, const char *conf)
        minimum size of the context => pass NULL to conf
        *tcti_context => desired connection properties for the TCTI
    */
    rc = Tss2_Tcti_Tabrmd_Init(nullptr, &context_size, NULL);
    if(!ErrorHandling(rc))
    {
        fprintf(stdout,"Failed to get allocation size for mssim TCTI \n");
        exit(EXIT_FAILURE);
    }
    
    tcti_context = (TSS2_TCTI_CONTEXT *) calloc(1,context_size);
    if(!ErrorHandling(rc))
    {
        fprintf(stdout, "Allocation for TCTI context failed: %s0 \n", strerror(errno));
        exit(EXIT_FAILURE);
    }

    rc = Tss2_Tcti_Tabrmd_Init(tcti_context,&context_size, NULL);
    if(!ErrorHandling(rc))
    {
        fprintf(stdout, "Failed to initialize tabrmd TCTI context: 0x%" PRIx32 "0\n", rc);
        free(tcti_context);
        exit(EXIT_FAILURE);
    }

    fprintf(stdout,"Initialization of the TCTI context successfull \n");
}

void pTPM::Init_Esys_Context()
{
    TSS2_RC rc;
    TSS2_ABI_VERSION abi_version = TSS2_ABI_VERSION_CURRENT;

    //Initialization of the ESYS context
    /*
        TSS2_RC Esys_Initialize( ESYS_CONTEXT **esysContext, TSS2_TCTI_CONTEXT *tcti, TSS2_ABI_VERSION *abiVersion);
        esysContext => reference to a pointer to the opaque ESYS_CONTEXT blob. It must be not NULL
        tcti => pointer to the TCTI context
        abiVersion => pointer to the ABI version that the application requests
    */
    rc = Esys_Initialize(&esys_context,tcti_context,&abi_version);
    if (rc != TSS2_RC_SUCCESS)
    {
        fprintf(stderr,"Failed to initialize the ESYS context: 0x%" PRIx32 "0 \n", rc);
        free(tcti_context);
        exit(EXIT_FAILURE);
    }

    fprintf(stdout,"Initialization of the ESYS context successfull \n");
}

void pTPM::Finalize_Tcti_Tabrmd_Context()
{
    Esys_Finalize(&esys_context);
    free(tcti_context);
    
    fprintf(stdout,"Finalization of the Esys context \n");
}

void pTPM::Finalize_Esys_Context()
{
    Tss2_Tcti_Finalize(tcti_context);

    fprintf(stdout,"Finalization of the TCTI context \n");
}

UINT16 pTPM::GetHashSize(TPMI_ALG_HASH hash)
{
    UINT16 hashSize = 0;

    switch (hash)
    {
    case TPM2_ALG_SHA1:
        hashSize = 20;
        break;
    case TPM2_ALG_SHA256:
        hashSize = 32;
        break;
    default:
        break;
    }

    return hashSize;
}

// Map Handling methods
void pTPM::UpdateMap(unordered_map<MapKey,ESYS_TR>* mapNVarea, UINT32 identifier, TPMI_ALG_HASH hash, UINT32 pcrNumber, ESYS_TR nvIndex)
{   
    MapKey mapKey(identifier,hash,pcrNumber);
    mapNVarea->insert(make_pair(mapKey,nvIndex));
}

ESYS_TR pTPM::GetIndexFromMap(UINT32 vID, TPMI_ALG_HASH hash, UINT32 pcrNumber)
{
	ESYS_TR returnValue = ESYS_TR_NONE;
    
    returnValue=mapNVarea.find(MapKey(vID,hash,pcrNumber))->second;
    
    if(returnValue!=ESYS_TR_NONE)
    {
        return returnValue;
    }
    else
    {
        fprintf(stderr,"Error during the Index identification\n");
        exit(EXIT_FAILURE);
    }

}
//////////////////////////////////////////////////////
///             ERROR HANDLING METHOD              ///
//////////////////////////////////////////////////////
int pTPM::ErrorHandling(TSS2_RC rc)
{
    switch (rc)
    {
    case TSS2_RC_SUCCESS :
        return 1;
    case TSS2_ESYS_RC_BAD_REFERENCE:
        fprintf(stderr,"Bad Reference: the esysContext or required input pointers or required output handle references are NULL(0x%" PRIx32 " )\n",rc);
        return 0;
    case TSS2_ESYS_RC_MEMORY:
        fprintf(stderr,"he ESAPI cannot allocate enough memory for internal operations or return parameters(0x%" PRIx32 " )\n",rc);
        return 0;
    case TSS2_ESYS_RC_BAD_SEQUENCE:
        fprintf(stderr,"the context has an asynchronous operation already pending(0x%" PRIx32 " )\n",rc);
        return 0;
    case TSS2_ESYS_RC_INSUFFICIENT_RESPONSE:
        fprintf(stderr,"the TPM's response does not at least contain the tag, response length, and response code(0x%" PRIx32 " )\n",rc);
        return 0;
    case TSS2_ESYS_RC_MALFORMED_RESPONSE:
        fprintf(stderr,"the TPM's response is corrupted(0x%" PRIx32 " )\n",rc);
        return 0;
    case TSS2_ESYS_RC_RSP_AUTH_FAILED:
        fprintf(stderr,"the response HMAC from the TPM did not verify(0x%" PRIx32 " )\n",rc);
        return 0;
    case TSS2_ESYS_RC_MULTIPLE_DECRYPT_SESSIONS:
        fprintf(stderr,"more than one session has the 'decrypt' attribute bit set(0x%" PRIx32 " )\n",rc);
        return 0;
    case TSS2_ESYS_RC_MULTIPLE_ENCRYPT_SESSIONS:
        fprintf(stderr,"more than one session has the 'decrypt' attribute bit set(0x%" PRIx32 " )\n",rc);
        return 0;
    case TSS2_ESYS_RC_BAD_TR:
        fprintf(stderr,"ny of the ESYS_TR objects are unknown to the ESYS_CONTEXT or are of the wrong type or if required ESYS_TR objects are ESYS_TR_NONE(0x%" PRIx32 " )\n",rc);
        return 0;
    case TSS2_ESYS_RC_NO_ENCRYPT_PARAM:
        fprintf(stderr,"one of the sessions has the 'encrypt' attribute set and the command does not support encryption of the first response parameter(0x%" PRIx32 " )\n",rc);
        return 0;
    case TPM2_RC_NV_UNINITIALIZED:
        fprintf(stdout,"Read of NV location before it is written: 0x%" PRIx32 "\n", rc);
        return 0;
    case TSS2_RESMGR_RC_LAYER | TSS2_RESMGR_TPM_RC_LAYER:
        fprintf(stderr,"Command TPM2_GetCommandAuditDigest not supported by TPM : 0x%" PRIx32 " \n",rc);
        return 0;
    default:
        fprintf(stderr,"Error number : 0x%" PRIx32 " \n",rc);
        return 0;
    }
}

/*/////////////////////////////////////////////////////////
                    GENERAL METHODS
/////////////////////////////////////////////////////////*/

void pTPM::Allocate_VM_Resources(UINT32 vID)
{
	UINT32 id = vID;
	// Return code
    TSS2_RC rc;

    // nvHandler
	ESYS_TR nvHandle = ESYS_TR_NONE;

    // Variable definition
    int i = 0;
    TPMI_ALG_HASH hashBank1 = TPM2_ALG_SHA1;
    TPMI_ALG_HASH hashBank256 = TPM2_ALG_SHA256;

    // All the PCRs for each bank will be allocated for the new VM that is deployed
    fprintf(stderr,"Memory allocation for BANK SHA1\n");
    for (i = 0; i < 3; i++)
    {
    	// Last index defined in the NVRAM
    	freeIndex +=1;

    	// Allocation of the nvPCR
    	nvHandle = Define_nvPCR(hashBank1,freeIndex);

    	// Updating the Map that is storing the memory location of each PCR
    	UpdateMap(&mapNVarea,id,hashBank1,i,nvHandle);
    }

    i = 0;

    fprintf(stderr,"Memory allocation for BANK SHA256\n");
    for (i = 0; i < 3; i++)
    {
    	// Last index defined in the NVRAM
    	freeIndex +=1;

    	// Allocation of the nvPCR
    	nvHandle = Define_nvPCR(hashBank256,freeIndex);

    	// Updating the Map that is storing the memory location of each PCR
    	UpdateMap(&mapNVarea,vID,hashBank256,i,nvHandle);
    }
}

void pTPM::Release_VM_Resources(UINT32 vID)
{
	// Return code
    TSS2_RC rc;

    // nvHandler
	ESYS_TR nvHandle = ESYS_TR_NONE;
	TPMI_ALG_HASH hashBank1 = TPM2_ALG_SHA1;
    TPMI_ALG_HASH hashBank256 = TPM2_ALG_SHA256;	

	// Variable Definition
	int i = 0;

	// Release resources that were reserved for a VM
    for (i = 0; i < 3; i++)
    {
    	// Retrieving the correct index from the Map 
    	nvHandle = GetIndexFromMap(vID, hashBank1, i);

    	// Releasing resources
    	Undefine_NV_Space(&nvHandle);
    }

   	i = 0;

	// Release resources that were reserved for a VM
    for (i = 0; i < 3; i++)
    {
    	// Retrieving the correct index from the Map 
    	nvHandle = GetIndexFromMap(vID, hashBank256, i);

    	// Releasing resources
    	Undefine_NV_Space(&nvHandle);
    }

}


ESYS_TR pTPM::Define_nvPCR (TPMI_ALG_HASH hashBank, TPMI_RH_NV_INDEX freeIndex)
{
	// Error handling
	TSS2_RC rc;

	// nvHandle for memory space that will be allocated
	ESYS_TR nvHandle = ESYS_TR_NONE;

	/*
        This structure is used for an authorization value and limits an authValue to being no larger than the
        largest digest produced by a TPM. In order to ensure consistency within an object, the authValue may be
        no larger than the size of the digest produced by the object’s nameAlg. This ensures that any TPM that
        can load the object will be able to handle the authValue of the object.
    */
    TPM2B_AUTH auth = {
        .size = 20,
        .buffer = {10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29}
    };

    // Definition of the public part
    TPM2B_NV_PUBLIC publicInfo = {
        .size = 0,
        .nvPublic = {
            .nvIndex = freeIndex,
            .nameAlg = hashBank,
            .attributes = (
                TPMA_NV_OWNERWRITE |
                TPMA_NV_AUTHWRITE |
                TPMA_NV_WRITE_STCLEAR |
                TPMA_NV_AUTHREAD |
                TPMA_NV_OWNERREAD |
                TPM2_NT_EXTEND << TPMA_NV_TPM2_NT_SHIFT     // TPM2_NT_EXTEND define the type of the index
                ),
            .authPolicy = {
            .size = 0,
            .buffer = {},
        },
        .dataSize = GetHashSize(hashBank)
        }
    };

    rc = Esys_NV_DefineSpace(
        esys_context,           // esysContext [in/out]
        ESYS_TR_RH_OWNER,       // [in] authHandle TPM2_RH_OWNER
        ESYS_TR_PASSWORD,       // [in] Session handle for authorization of authHandle
        ESYS_TR_NONE,           // [in] Second session handle
        ESYS_TR_NONE,           // [in] Third session handle
        &auth,                  // [in] auth The authorization value
        &publicInfo,            // [in] publicInfo The public parameter of the NV area
        &nvHandle               // [out] nvHandle => ESYS_TR handle of ESYS resource for TPM2_HANDLE
    );

    // Checking if the definition of the space is done or not
    if( !ErrorHandling(rc))
    {
        fprintf(stdout,"Error during the definition of the space \n");
        exit(EXIT_FAILURE);
    } 
    else
    {
        fprintf(stdout,"Extend NV Index defined: ");
    }
    fprintf(stdout,"%u\n",nvHandle);

    return nvHandle;
}

void pTPM::Undefine_NV_Space(ESYS_TR* handle)
{
	TSS2_RC rc;

	// Undefinition of the NV space 
    rc = Esys_NV_UndefineSpace(
        esys_context,
        ESYS_TR_RH_OWNER,
        *handle,
        ESYS_TR_PASSWORD,
        ESYS_TR_NONE,ESYS_TR_NONE
    );

    // Error handling
    if(!ErrorHandling(rc))
    {
        fprintf(stdout,"Error during the Undefine procedure \n");
        exit(EXIT_FAILURE);
    }
}

void pTPM::Extend_nvPCR(UINT32 vID, int pcrNumber, TPMI_ALG_HASH hashBank, TPM2B_MAX_NV_BUFFER* dataToExtend)
{
	// Return code
    TSS2_RC rc;
    ESYS_TR nvHandle;
    
    nvHandle = GetIndexFromMap(vID, hashBank, pcrNumber);
    cout << dataToExtend->buffer << endl;

    /*
        If the index type is of type Extend, the only way to modify it is with Extend command
        TPM extension function:
        nvIndex→datanew ≔ HnameAlg (nvIndex→data.old || data.buffer)
    */
    rc = Esys_NV_Extend(
        esys_context,       // [in] esysContext
        nvHandle,           // [in] authHandle indicating the source of the authorization handle
        nvHandle,           // [in] nvIndex The nv index to extend
        ESYS_TR_PASSWORD,   // [in] Session handle for the authorization handle
        ESYS_TR_NONE,       // [in] Second session handle
        ESYS_TR_NONE,       // [in] Third session handle
        dataToExtend        // [in] data The data to extend
    );

    if(!ErrorHandling(rc))
    {
        fprintf(stdout,"Error during the extend operation \n");
        exit(EXIT_FAILURE);
    } 
    else
    {
        fprintf(stdout,"Extension of the space %u successfull \n",nvHandle);
    }
}

void pTPM::CreatePK()
{
    // Return code 
    TSS2_RC rc;

    ESYS_TR primaryHandle = ESYS_TR_NONE;
    ESYS_TR primaryHandleAuthSession = ESYS_TR_NONE;
    
    // Approximation : no session or HMAC authorization defined
    TPM2B_AUTH authValuePK = {
        .size = 5,
        .buffer = {1,2,3,4,5}
    };

    TPM2B_SENSITIVE_CREATE inSensitivePK = {
        .size = 0,
        .sensitive = {
            .userAuth = authValuePK,
            .data = {
                .size = 0,
                .buffer = {0}
            },
        },
    };

    TPM2B_PUBLIC inPublic = {
        .size = 0,
        .publicArea = {
            .type = TPM2_ALG_RSA,
            .nameAlg = TPM2_ALG_SHA256,
            .objectAttributes = (
                TPMA_OBJECT_USERWITHAUTH | 
                TPMA_OBJECT_RESTRICTED |
                TPMA_OBJECT_DECRYPT |
                TPMA_OBJECT_FIXEDTPM |
                TPMA_OBJECT_FIXEDPARENT |
                TPMA_OBJECT_SENSITIVEDATAORIGIN 
            ),
            .authPolicy = {
                .size = 0,
                .buffer = {},
            },
            .parameters = {
                .rsaDetail = {
                    .symmetric = {
                        .algorithm = TPM2_ALG_AES,
                        .keyBits = {
                            .aes = 128
                        },
                        .mode = {
                            .aes = TPM2_ALG_CFB,
                        },
                    },
                    .scheme = {
                        .scheme = TPM2_ALG_NULL,
                    /*    .details = {
                            .rsapss = {
                                .hashAlg = TPM2_ALG_SHA256
                            }
                        }*/
                    },
                    .keyBits = 2048,
                    .exponent = 0,
                },
            },
            .unique = {
                .rsa = {
                    .size = 0,
                    .buffer = {},
                },
            },
        },
    };

    TPM2B_DATA outInfo = {.size = 0, .buffer = {}};
    TPML_PCR_SELECTION pcrSelection = {.count = 0};
    TPM2B_AUTH authValue = {.size = 0, .buffer = {}};

    rc = Esys_TR_SetAuth(esys_context,ESYS_TR_RH_OWNER,&authValue);

    TPM2B_PUBLIC *outPublic;
    TPM2B_CREATION_DATA *creationData;
    TPM2B_DIGEST *creationHash;
    TPMT_TK_CREATION *creationTicket;

    rc = Esys_CreatePrimary(
        esys_context,           // [in] esysContext
        ESYS_TR_RH_OWNER,       // [in] primaryHandle : hierarchy
        ESYS_TR_PASSWORD,       // [in] Session handle for authorization of primaryHandle
        ESYS_TR_NONE,           // Session handle 2
        ESYS_TR_NONE,           // Session handle 3
        &inSensitivePK,         // [in] inSensitive => sensitive data
        &inPublic,              // [in] public Template
        &outInfo,               // [in] data that will be included in the creation data this object to provide permanent
        &pcrSelection,          // [in] PCR that will be used in the creation process
        &primaryHandle,         // [in] object handle
        &outPublic,             // [out] Public portion of the created object 
        &creationData,          // [out] contains the creation data
        &creationHash,          // [out] Digest of creation data
        &creationTicket         // [out] ticket used to validate the creation of the object
    );

    if(!ErrorHandling(rc))
    {
        fprintf(stdout,"Error during the creation of the primary key\n");
        exit(EXIT_FAILURE);
    } else
    {
        fprintf(stdout,"Creation of the Key successfull \n");
    }

    rc = Esys_TR_SetAuth(esys_context,primaryHandle,&authValuePK);

    primaryKey.keyHandle = primaryHandle;
    primaryKey.creationData = creationData;
    primaryKey.creationHash = creationHash;
    primaryKey.creationTicket = creationTicket;
    primaryKey.outPublic = outPublic;
}

void pTPM::Create_Key(Key* derivedKey, Key* primaryKey)
{
    // Return code 
    TSS2_RC rc;

    // Starting an authorization session
    ESYS_TR session = ESYS_TR_NONE;
    ESYS_TR primaryHandleAuthSession = primaryKey->keyHandle;
    TPMT_SYM_DEF symmetric = {.algorithm = TPM2_ALG_AES,
                              .keyBits = {.aes = 128},
                              .mode = {.aes = TPM2_ALG_CFB}};
    TPMA_SESSION sessionAttributes;
    sessionAttributes = (TPMA_SESSION_DECRYPT |
                         TPMA_SESSION_ENCRYPT |
                         TPMA_SESSION_CONTINUESESSION);
    TPM2_SE sessionType = TPM2_SE_HMAC;
    TPMI_ALG_HASH authHash = TPM2_ALG_SHA256;

    rc = Esys_StartAuthSession(
        esys_context,
        primaryHandleAuthSession,
        ESYS_TR_NONE,
        ESYS_TR_NONE,
        ESYS_TR_NONE,
        ESYS_TR_NONE,
        NULL,
        sessionType,
        &symmetric,
        authHash,
        &session
    );

    rc = Esys_TRSess_SetAttributes(
        esys_context,
        session,
        sessionAttributes,
        0xff
    );
                         
    // Creation of the derived key
    TPM2B_AUTH authDerivedKey = {
        .size = 5,
        .buffer = {1,2,3,4,5}
    };

    TPM2B_SENSITIVE_CREATE inSensitiveDK = {
        .size = 0,
        .sensitive = {
            .userAuth = {
                .size = 0,
                .buffer = {0}
            },
            .data = {
                .size=0,
                .buffer = {0}
            }
        }
    };

    inSensitiveDK.sensitive.userAuth = authDerivedKey;

    TPM2B_PUBLIC inPublicDK = {
        .size = 0,
        .publicArea = {
            .type = TPM2_ALG_RSA,
            .nameAlg = TPM2_ALG_SHA256,
            .objectAttributes = (
                TPMA_OBJECT_USERWITHAUTH | 
                TPMA_OBJECT_SIGN_ENCRYPT |
                TPMA_OBJECT_FIXEDTPM |
                TPMA_OBJECT_FIXEDPARENT |
                TPMA_OBJECT_SENSITIVEDATAORIGIN
            ),
            .authPolicy = {
                .size = 0,
                .buffer = {},
            },
            .parameters = {
                .rsaDetail = {
                    .symmetric = {
                        .algorithm = TPM2_ALG_NULL,
                        .keyBits = {
                            .aes = 128
                        },
                        .mode = {
                            .aes = TPM2_ALG_CFB,
                        },
                    },
                    .scheme = {
                        .scheme = TPM2_ALG_RSAPSS,
                        .details = {
                            .rsapss = {
                                .hashAlg = TPM2_ALG_SHA256
                            }
                        }
                    },
                    .keyBits = 2048,
                    .exponent = 0,
                },
            },
            .unique = {
                .rsa = {
                    .size = 0,
                    .buffer = {},
                },
            },
        },
    };

    TPM2B_DATA outInfoDK = {.size = 0, .buffer = {}};
    TPML_PCR_SELECTION creationPCRdk = { .count = 0,};

    TPM2B_PUBLIC *outPublicDerived;
    TPM2B_PRIVATE *outPrivateDerived;
    TPM2B_CREATION_DATA *creationDataDerived;
    TPM2B_DIGEST *creationHashDerived;
    TPMT_TK_CREATION *creationTicketDerived;

    rc = Esys_Create(
        esys_context,
        primaryKey->keyHandle,
        session,
        ESYS_TR_NONE,
        ESYS_TR_NONE,
        &inSensitiveDK,
        &inPublicDK,
        &outInfoDK,
        &creationPCRdk,
        &outPrivateDerived,
        &outPublicDerived,
        &creationDataDerived,
        &creationHashDerived,
        &creationTicketDerived
    );
    
    if(!ErrorHandling(rc))
    {
        fprintf(stdout,"Error during the creation of the derived key\n");
        exit(EXIT_FAILURE);
    } else
    {
        fprintf(stdout,"Creation of the derived key successfull \n");
    }

    derivedKey->keyHandle = ESYS_TR_NONE;
    derivedKey->creationData = creationDataDerived;
    derivedKey->creationHash = creationHashDerived;
    derivedKey->creationTicket = creationTicketDerived;
    derivedKey->outPrivate = outPrivateDerived;
    derivedKey->outPublic = outPublicDerived;
}

void pTPM::CreateAIK(UINT32 vID)
{
	Create_Key(&keyVM1, &primaryKey);
}

IntegrityReport pTPM::IntegrityRequestFrom_vTPM(UINT32 vID, IntegrityRequest integrityRequest)
{
	IntegrityReport integrityReport;

	// Extracting from the Integrity request the useful data
	cout << "Extracting Integrity Request parameters" << endl;
	TPM2B_DATA nonce = integrityRequest.nonce;
	TPML_PCR_SELECTION pcrSelection = integrityRequest.pcrSelection;
	
	TPMT_SIG_SCHEME inScheme = {.scheme = TPM2_ALG_NULL};
	TPM2B_ATTEST attestStructure={.size=135};
	TPMT_SIGNATURE signatureStructure;
	TPML_DIGEST pcrValues= {
        .count = pcrSelection.count
        };
    TPM2B_ATTEST attestStructure_VMM={.size=135};
	TPMT_SIGNATURE signatureStructure_VMM;
	TPML_DIGEST pcrValues_VMM= {
        .count = pcrSelection.count
        };
    cout << "Permorming quote operation" << endl ;

	TPM2_VIRT_QUOTE(
		vID,
		pcrSelection,
		nonce,
		inScheme,
		&keyVM1,
		&primaryKey,
		&attestStructure,
		&signatureStructure,
		&pcrValues
		);

	integrityReport.attestStructure_VM = attestStructure;
	integrityReport.signatureStructure_VM = signatureStructure;
	integrityReport.pcrValues_VM = pcrValues;

    TPM2_VIRT_QUOTE(
		0,
		pcrSelection,
		nonce,
		inScheme,
		&keyVM1,
		&primaryKey,
		&attestStructure_VMM,
		&signatureStructure_VMM,
		&pcrValues_VMM
		);

	integrityReport.attestStructure_VMM = attestStructure_VMM;
	integrityReport.signatureStructure_VMM = signatureStructure_VMM;
	integrityReport.pcrValues_VMM = pcrValues_VMM;

	return integrityReport ;
}

void pTPM::TPM2_VIRT_QUOTE(UINT32 vID, TPML_PCR_SELECTION pcrSelection, TPM2B_DATA nonce, TPMT_SIG_SCHEME inScheme, Key* signingKey, Key* primaryKey, TPM2B_ATTEST* attestStructure, TPMT_SIGNATURE* signature, TPML_DIGEST* pcrValues)
{
	//Return code
    ESYS_TR rc;

    TPMT_SIGNATURE* signatureStructure;

    // Loading the signing key
    ESYS_TR signKeyHandle;
    ESYS_TR session = ESYS_TR_NONE;
    ESYS_TR primaryHandleAuthSession = primaryKey->keyHandle;

    // This structure is used to select an algorithm to be used for parameter encryption
    TPMT_SYM_DEF symmetric = {.algorithm = TPM2_ALG_AES,
                              .keyBits = {
                                  .aes = 128
                                  },
                              .mode = {
                                  .aes = TPM2_ALG_CFB
                                  }
                            };

    //Used to select session attributes
    TPMA_SESSION sessionAttributes; 
    sessionAttributes = (TPMA_SESSION_DECRYPT | // This indicates that the command is encrypted 
                         TPMA_SESSION_ENCRYPT | // To use the session to encrypt
                         TPMA_SESSION_CONTINUESESSION); // The session remain active after successfull completion of the command
    TPM2_SE sessionType = TPM2_SE_HMAC;
    TPMI_ALG_HASH authHash = TPM2_ALG_SHA256;

    rc = Esys_StartAuthSession(
        esys_context,
        primaryHandleAuthSession,
        ESYS_TR_NONE,
        ESYS_TR_NONE,
        ESYS_TR_NONE,
        ESYS_TR_NONE,
        NULL,
        sessionType,
        &symmetric,
        authHash,
        &session
    );

    rc = Esys_TRSess_SetAttributes(
        esys_context,
        session,
        sessionAttributes,
        0xff
    );

    // Loading the signing key
    rc = Esys_Load(
        esys_context,
        primaryKey->keyHandle,
        session,
        ESYS_TR_NONE,
        ESYS_TR_NONE,
        signingKey->outPrivate,
        signingKey->outPublic,
        &signKeyHandle
    );

    TPM2B_AUTH authKey = {
        .size = 5,
        .buffer = {1,2,3,4,5}
    };

    rc = Esys_TR_SetAuth(esys_context,signKeyHandle, &authKey);

    //Extracting nvIndex
    int nvEnd = pcrSelection.count;
    ESYS_TR nvIndex;
    UINT32 pcrNumber;
    TPMI_ALG_HASH hash = TPM2_ALG_SHA1;

    // Creating an array to store the nvIndex handles corresponding to the different PCR requested
    ESYS_TR arrayNVhandle[pcrSelection.count];

    // If the PCRs selected come from different hash banks => array to store the correspondign hash 
    TPMI_ALG_HASH hashSelection[pcrSelection.count];

    // With this routine the PCR number to be attested are extrapolated from the TPML_PCR_SELEECTION structure,
    cout<< endl << "Extrapolation of the vPCR"<<endl;

    // which is the structure used in the TPM2_Quote()
    for (int i = 0; i < pcrSelection.count; i++)
    {
        // Useful for searching the corresponding handles 
        hash = pcrSelection.pcrSelections[i].hash;
        
        for (int j = 0; j <pcrSelection.pcrSelections[i].sizeofSelect; j++)
        {
            if(pcrSelection.pcrSelections[i].pcrSelect[j]!=0)
            {
                // Extrapolation from the bitmap
                pcrNumber = (j*8)+log2(pcrSelection.pcrSelections[i].pcrSelect[j]);
                
                // Populating the arrays
                arrayNVhandle[i] = GetIndexFromMap(vID,hash,pcrNumber);
                hashSelection[i] = hash;
            }
        }
    }

    // Determining the size of the final digest, it is based on the algorithm used for the different PCRs
    UINT16 sizeOfDigest=0 ;
    for (int i=0; i<pcrSelection.count;i++)
    {
        if(hashSelection[i]==TPM2_ALG_SHA1)
        {
            sizeOfDigest+=20;
            pcrValues->digests[i].size = 20; // setting the size of the digest for the response
            //fprintf(stdout,"Size of the digest: %u\n",pcrValues.digests[i].size);
        }
        else if(hashSelection[i]==TPM2_ALG_SHA256)
        {
            sizeOfDigest+=32;
            pcrValues->digests[i].size = 32; // setting the size of the digest for the response
            //fprintf(stdout,"Size of the digest: %u\n",pcrValues.digests[i].size);
        }
    }
    fprintf(stdout,"Size of the index to be defined: %u\n",sizeOfDigest);


    TPM2B_MAX_BUFFER pcrConcatenation = {
        .size = sizeOfDigest,
        .buffer = {0}
    };

    // Populating the the buffer with the concatenation of all the digests
    TPM2B_MAX_NV_BUFFER* readPCR;
    int index_start=0;
    int index_end;

    cout<<"pcrConcatenation"<<endl;

    size_t index_offset = 0;

    for (int i = 0; i < pcrSelection.count; i++)
    {
        rc = Esys_NV_Read(
            esys_context,
            arrayNVhandle[i],
            arrayNVhandle[i],
            ESYS_TR_PASSWORD,
            ESYS_TR_NONE,
            ESYS_TR_NONE,
            GetHashSize(hashSelection[i]),
            0,
            &readPCR
            );

        for (int j = index_start; j < index_start+readPCR->size; j++)
        {
            pcrConcatenation.buffer[j]=readPCR->buffer[j-index_start];
            // Saving the PCR values to be included in the quote answer
            pcrValues->digests[i].buffer[j-index_start]=readPCR->buffer[j-index_start];
            fprintf(stdout, "%" PRIX8 " ",pcrConcatenation.buffer[j]);
        }
        index_start+=readPCR->size;
    }
    cout << endl;

    //Creation of the hash for the quote
    cout<<"Digest creation"<<endl;

    TPM2B_DIGEST* outHash;
    TPMT_TK_HASHCHECK* hashValidation;
    TPMI_ALG_HASH hashAlgorithm = signingKey->outPublic->publicArea.nameAlg;

    rc = Esys_Hash(
        esys_context,
        ESYS_TR_NONE,
        ESYS_TR_NONE,
        ESYS_TR_NONE,
        &pcrConcatenation,
        hashAlgorithm,
        TPM2_RH_NULL,
        &outHash,
        &hashValidation
    );

    for (int i = 0; i < outHash->size; i++)
    {
        fprintf(stdout, "%" PRIX8 "", outHash->buffer[i] );
    }

    // Creation of of the structure to be signed
    size_t offset = 0;

    //attestStructure.attestationData magic = TPM2_GENERATED_VALUE;
    Tss2_MU_INT32_Marshal(
        TPM2_GENERATED_VALUE,
        attestStructure->attestationData,
        attestStructure->size,
        &offset);

    cout<<"\n Magic Number \n"<<endl;
    //fprintf(stdout, "%zu\n", offset);

    for (size_t i = 0; i < offset; i++)
    {
        fprintf(stdout, "%" PRIX8 "", attestStructure->attestationData[i] );
    }

    // type TPMI_ST_ATTEST = TPMI_ST_ATTEST_QUOTE
    Tss2_MU_INT16_Marshal(
        TPM2_ST_ATTEST_QUOTE,
        attestStructure->attestationData,
        attestStructure->size,
        &offset);

    cout<<"\n Type \n"<<endl;
    //fprintf(stdout, "%zu\n", offset);

    for (size_t i = 0; i < offset; i++)
    {
        fprintf(stdout, "%" PRIX8 "", attestStructure->attestationData[i] );
    }

    // Qualified Signer Name = We can get QN from Esys_Readpublic => QN of the signging key

    TPM2B_PUBLIC* outPublic;
    TPM2B_NAME* name;
    TPM2B_NAME* qualifiedName;

    rc = Esys_ReadPublic(
        esys_context,
        signKeyHandle,
        session,
        ESYS_TR_NONE,
        ESYS_TR_NONE,
        &outPublic,
        &name,
        &qualifiedName
        );

    cout<<"\nQN signing key"<<endl;
    for (int i = 0; i < qualifiedName->size ; i++)
    {
        fprintf(stdout, "%" PRIX8 "", qualifiedName->name[i] );
    }

    Tss2_MU_TPM2B_NAME_Marshal(
        qualifiedName,
        attestStructure->attestationData,
        attestStructure->size,
        &offset);

    cout<<"\n Extra Data \n"<<endl;
    //fprintf(stdout, "%zu\n", offset);
    for (size_t i = 0; i < offset; i++)
    {
        fprintf(stdout, "%" PRIX8 "", attestStructure->attestationData[i] );
    };

    // Extra Data
    TPM2B_DATA extraData = nonce;

    Tss2_MU_TPM2B_DATA_Marshal(
        &extraData,
        attestStructure->attestationData,
        attestStructure->size,
        &offset);

    cout<<"\n Extra Data \n"<<endl;
    //fprintf(stdout, "%zu\n", offset);
    for (size_t i = 0; i < offset; i++)
    {
        fprintf(stdout, "%" PRIX8 "", attestStructure->attestationData[i] );
    }

    // Clock Info 
    TPMS_TIME_INFO *currentTime;
    TPMS_CLOCK_INFO clockInfo;
    rc = Esys_ReadClock(
        esys_context,
        ESYS_TR_NONE,
        ESYS_TR_NONE,
        ESYS_TR_NONE,
        &currentTime
        );

    clockInfo = currentTime->clockInfo;

    Tss2_MU_TPMS_CLOCK_INFO_Marshal(
        &clockInfo,
        attestStructure->attestationData,
        attestStructure->size,
        &offset);

    cout<<"\n Clock Info \n"<<endl;
    //fprintf(stdout, "%zu\n", offset);

    for (size_t i = 0; i < offset; i++)
    {
        fprintf(stdout, "%" PRIX8 "", attestStructure->attestationData[i] );
    }

    // Firmware Version
    Tss2_MU_UINT32_Marshal(
        TPM2_PT_FIRMWARE_VERSION_1,
        attestStructure->attestationData,
        attestStructure->size,
        &offset);

    Tss2_MU_UINT32_Marshal(
       TPM2_PT_FIRMWARE_VERSION_2,
       attestStructure->attestationData,
       attestStructure->size,
       &offset);

    cout<<"\n Firmware Version \n"<<endl;
    //fprintf(stdout, "%zu\n", offset);
    for (size_t i = 0; i < offset; i++)
    {
        fprintf(stdout, "%" PRIX8 "", attestStructure->attestationData[i] );
    }

    // attested TPMU_ATTEST
    TPMU_ATTEST attested;
    attested.quote.pcrSelect = pcrSelection;
    attested.quote.pcrDigest = *outHash;

    Tss2_MU_TPMU_ATTEST_Marshal(
        &attested,
        TPM2_ST_ATTEST_QUOTE,
        attestStructure->attestationData,
        attestStructure->size,
        &offset);

    cout<<"\n TPMU_ATTEST \n"<<endl;
    //fprintf(stdout, "%zu\n", offset);
    for (size_t i = 0; i < offset; i++)
    {
        fprintf(stdout, "%" PRIX8 "", attestStructure->attestationData[i] );
    }
    cout << endl;
    for (size_t i = 0; i < attestStructure->size; i++)
    {
        fprintf(stdout, "%" PRIX8 "", attestStructure->attestationData[i] );
    }
    cout << endl<< offset;

    // Now that we have created the structure we need to calculate the hash in order to sign it
    TPM2B_DIGEST* quoteHash;
    TPMT_TK_HASHCHECK* quoteHashValidation;
    TPM2B_MAX_BUFFER bufferToHash = {.size=attestStructure->size, .buffer={0}};
    offset = 0;

    // Copying the attestStructure in a suitable structure to be hashed
    for (int i = 0; i < attestStructure->size; ++i)
    {
        bufferToHash.buffer[i]=attestStructure->attestationData[i];
    }

    cout<<"bufferToHash"<<endl;
    for (int i = 0; i < bufferToHash.size ; i++)
    {
        fprintf(stdout, "%" PRIX8 "", bufferToHash.buffer[i] );
    }

    cout<<"\nAttestStructure"<<endl;
    for (int i = 0; i < attestStructure->size ; i++)
    {
        fprintf(stdout, "%" PRIX8 "", attestStructure->attestationData[i] );
    }

/*
    We create the hash of the structure that will be signed
*/
    rc = Esys_Hash(
        esys_context,
        ESYS_TR_NONE,
        ESYS_TR_NONE,
        ESYS_TR_NONE,
        &bufferToHash,
        hashAlgorithm, // to be taken from signing scheme
        TPM2_RH_NULL,
        &quoteHash,
        &quoteHashValidation
    );

    cout<<"\nHashed structure"<<endl;
    for (int i = 0; i < quoteHash->size ; i++)
    {
        fprintf(stdout, "%" PRIX8 "", quoteHash->buffer[i] );
    }

    // Structure Created , time to sign the structure as it is done in the quote operation 
    rc = Esys_Sign(
        esys_context,
        signKeyHandle,
        session,
        ESYS_TR_NONE,
        ESYS_TR_NONE,
        quoteHash,
        &inScheme,
        quoteHashValidation,
        &signatureStructure);

    fprintf(stdout,"\nSignature Created\n");
    cout << signatureStructure->signature.rsapss.sig.size << endl ;
    for (int i = 0; i < signatureStructure->signature.rsapss.sig.size ; i++)
    {
        fprintf(stdout,"%" PRIX8 " ",signatureStructure->signature.rsapss.sig.buffer[i]);
    }

    *signature = *signatureStructure;

/*
    TPMT_TK_VERIFIED *validationSignature;
    rc = Esys_VerifySignature(
        esys_context,
        signKeyHandle,
        ESYS_TR_NONE,
        ESYS_TR_NONE,
        ESYS_TR_NONE,
        quoteHash,
        signatureStructure,
        &validationSignature
    );

    if(!ErrorHandling(rc))
    {
        fprintf(stdout,"Error during signature verification\n");
    } else
    {
        fprintf(stdout,"Signature Verified \n");
    }
    */
}

Key pTPM::GetRootKey()
{
	return primaryKey;
}

Key pTPM::GetAIK()
{
	return keyVM1;
}

Key* pTPM::GetRootKeyP()
{
	return &primaryKey;
}

Key* pTPM::GetAIKP()
{
	return &keyVM1;
}

