#include "vTPM.h"
#include "tss2/tss2_common.h"
#include "tss2/tss2_tpm2_types.h"
#include "tss2/tss2-tcti-tabrmd.h"
#include "stdio.h"
#include "stdlib.h"
#include <inttypes.h>
#include <string.h>
#include <string>
#include <fstream>
#include <errno.h>



// Constructor
vTPM::vTPM(UINT32 id)
{
    identifier = id;
    nvHandle = ESYS_TR_NONE;

    // Initialization of the contexts
    Init_Tcti_Tabrmd_Context();
    Init_Esys_Context();

    fprintf(stdout,"TPM instantiation successfull \n");
}

// Destructor
vTPM::~vTPM()
{

}

void vTPM::Init_Tcti_Tabrmd_Context()
{
    TSS2_RC rc;
    size_t context_size;

    //Define allocation for the mssim TCTI
    /* 
        TSS2_RC Tss2_Tcti_Tabrmd_Init (TSS2_TCTI_CONTEXT *tcti_context, size_t *size, const char *conf)
        minimum size of the context => pass NULL to conf
        *tcti_context => desired connection properties for the TCTI
    */

    rc = Tss2_Tcti_Tabrmd_Init(nullptr, &context_size, NULL);
    if(!ErrorHandling(rc))
    {
        fprintf(stdout,"Failed to get allocation size for mssim TCTI \n");
        exit(EXIT_FAILURE);
    }
    
    tcti_context = (TSS2_TCTI_CONTEXT *) calloc(1,context_size);
    if(!ErrorHandling(rc))
    {
        fprintf(stdout, "Allocation for TCTI context failed: %s0 \n", strerror(errno));
        exit(EXIT_FAILURE);
    }

    rc = Tss2_Tcti_Tabrmd_Init(tcti_context,&context_size, NULL);
    if(!ErrorHandling(rc))
    {
        fprintf(stdout, "Failed to initialize tabrmd TCTI context: 0x%" PRIx32 "0\n", rc);
        free(tcti_context);
        exit(EXIT_FAILURE);
    }

    fprintf(stdout,"Initialization of the TCTI context successfull \n");
}

void vTPM::Init_Esys_Context()
{
    TSS2_RC rc;
    TSS2_ABI_VERSION abi_version = TSS2_ABI_VERSION_CURRENT;

    //Initialization of the ESYS context
    /*
        TSS2_RC Esys_Initialize( ESYS_CONTEXT **esysContext, TSS2_TCTI_CONTEXT *tcti, TSS2_ABI_VERSION *abiVersion);
        esysContext => reference to a pointer to the opaque ESYS_CONTEXT blob. It must be not NULL
        tcti => pointer to the TCTI context
        abiVersion => pointer to the ABI version that the application requests
    */
    rc = Esys_Initialize(&esys_context,tcti_context,&abi_version);
    if (rc != TSS2_RC_SUCCESS)
    {
        fprintf(stderr,"Failed to initialize the ESYS context: 0x%" PRIx32 "0 \n", rc);
        free(tcti_context);
        exit(EXIT_FAILURE);
    }

    fprintf(stdout,"Initialization of the ESYS context successfull \n");
}

void vTPM::Finalize_Tcti_Tabrmd_Context()
{
    Esys_Finalize(&esys_context);
    free(tcti_context);
    
    fprintf(stdout,"Finalization of the Esys context \n");
}

void vTPM::Finalize_Esys_Context()
{
    Tss2_Tcti_Finalize(tcti_context);

    fprintf(stdout,"Finalization of the TCTI context \n");
}

//////////////////////////////////////////////////////
///             ERROR HANDLING METHOD              ///
//////////////////////////////////////////////////////

int vTPM::ErrorHandling(TSS2_RC rc)
{
    switch (rc)
    {
    case TSS2_RC_SUCCESS :
        return 1;
    case TSS2_ESYS_RC_BAD_REFERENCE:
        fprintf(stderr,"Bad Reference: the esysContext or required input pointers or required output handle references are NULL(0x%" PRIx32 " )\n",rc);
        return 0;
    case TSS2_ESYS_RC_MEMORY:
        fprintf(stderr,"he ESAPI cannot allocate enough memory for internal operations or return parameters(0x%" PRIx32 " )\n",rc);
        return 0;
    case TSS2_ESYS_RC_BAD_SEQUENCE:
        fprintf(stderr,"the context has an asynchronous operation already pending(0x%" PRIx32 " )\n",rc);
        return 0;
    case TSS2_ESYS_RC_INSUFFICIENT_RESPONSE:
        fprintf(stderr,"the TPM's response does not at least contain the tag, response length, and response code(0x%" PRIx32 " )\n",rc);
        return 0;
    case TSS2_ESYS_RC_MALFORMED_RESPONSE:
        fprintf(stderr,"the TPM's response is corrupted(0x%" PRIx32 " )\n",rc);
        return 0;
    case TSS2_ESYS_RC_RSP_AUTH_FAILED:
        fprintf(stderr,"the response HMAC from the TPM did not verify(0x%" PRIx32 " )\n",rc);
        return 0;
    case TSS2_ESYS_RC_MULTIPLE_DECRYPT_SESSIONS:
        fprintf(stderr,"more than one session has the 'decrypt' attribute bit set(0x%" PRIx32 " )\n",rc);
        return 0;
    case TSS2_ESYS_RC_MULTIPLE_ENCRYPT_SESSIONS:
        fprintf(stderr,"more than one session has the 'decrypt' attribute bit set(0x%" PRIx32 " )\n",rc);
        return 0;
    case TSS2_ESYS_RC_BAD_TR:
        fprintf(stderr,"ny of the ESYS_TR objects are unknown to the ESYS_CONTEXT or are of the wrong type or if required ESYS_TR objects are ESYS_TR_NONE(0x%" PRIx32 " )\n",rc);
        return 0;
    case TSS2_ESYS_RC_NO_ENCRYPT_PARAM:
        fprintf(stderr,"one of the sessions has the 'encrypt' attribute set and the command does not support encryption of the first response parameter(0x%" PRIx32 " )\n",rc);
        return 0;
    case TPM2_RC_NV_UNINITIALIZED:
        fprintf(stdout,"Read of NV location before it is written: 0x%" PRIx32 "\n", rc);
        return 0;
    case TSS2_RESMGR_RC_LAYER | TSS2_RESMGR_TPM_RC_LAYER:
        fprintf(stderr,"Command TPM2_GetCommandAuditDigest not supported by TPM : 0x%" PRIx32 " \n",rc);
        return 0;
    default:
        fprintf(stderr,"Error number : 0x%" PRIx32 " \n",rc);
        return 0;
    }
}

// Setter for handle
void vTPM::setNvHandle(ESYS_TR handle)
{
    this->nvHandle=handle;
}

// Getter for handle
ESYS_TR* vTPM::getNvHandle()
{
    return &nvHandle;
}

// Getter for identifier
UINT32 vTPM::getIdentifier()
{
    return identifier;
}

// Function used to load an external public key for the signature verification
ESYS_TR vTPM::Load_Public_Key(TPM2B_PUBLIC publicData)
{
    TSS2_RC rc;
    ESYS_TR publiKeyHandle;

    rc = Esys_LoadExternal(
        esys_context,
        ESYS_TR_NONE,
        ESYS_TR_NONE,
        ESYS_TR_NONE,
        NULL,
        &publicData,
        TPM2_RH_NULL,
        &publiKeyHandle
    );

    if(!ErrorHandling(rc))
    {
        fprintf(stdout,"Error during the load operation of the Public Key\n");
        exit(EXIT_FAILURE);
    } else
    {
        fprintf(stdout,"Virtual TPM loaded public key \n");
    }

    return publiKeyHandle;
}


// Method used to create a Primary key to sign the quoted PCRs
void vTPM::Create_PK(Key* primaryKey)
{
    // Return code 
    TSS2_RC rc;

    ESYS_TR primaryHandle = ESYS_TR_NONE;
    ESYS_TR primaryHandleAuthSession = ESYS_TR_NONE;
    
    // Approximation : no session or HMAC authorization defined
    TPM2B_AUTH authValuePK = {
        .size = 5,
        .buffer = {1,2,3,4,5}
    };

    TPM2B_SENSITIVE_CREATE inSensitivePK = {
        .size = 0,
        .sensitive = {
            .userAuth = authValuePK,
            .data = {
                .size = 0,
                .buffer = {0}
            },
        },
    };

    TPM2B_PUBLIC inPublic = {
        .size = 0,
        .publicArea = {
            .type = TPM2_ALG_RSA,
            .nameAlg = TPM2_ALG_SHA256,
            .objectAttributes = (
                TPMA_OBJECT_USERWITHAUTH | 
                TPMA_OBJECT_RESTRICTED |
                TPMA_OBJECT_DECRYPT |
                TPMA_OBJECT_FIXEDTPM |
                TPMA_OBJECT_FIXEDPARENT |
                TPMA_OBJECT_SENSITIVEDATAORIGIN 
            ),
            .authPolicy = {
                .size = 0,
                .buffer = {},
            },
            .parameters = {
                .rsaDetail = {
                    .symmetric = {
                        .algorithm = TPM2_ALG_AES,
                        .keyBits = {
                            .aes = 128
                        },
                        .mode = {
                            .aes = TPM2_ALG_CFB,
                        },
                    },
                    .scheme = {
                        .scheme = TPM2_ALG_NULL,
                    /*    .details = {
                            .rsapss = {
                                .hashAlg = TPM2_ALG_SHA256
                            }
                        }*/
                    },
                    .keyBits = 2048,
                    .exponent = 0,
                },
            },
            .unique = {
                .rsa = {
                    .size = 0,
                    .buffer = {},
                },
            },
        },
    };

    TPM2B_DATA outInfo = {.size = 0, .buffer = {}};
    TPML_PCR_SELECTION pcrSelection = {.count = 0};
    TPM2B_AUTH authValue = {.size = 0, .buffer = {}};

    rc = Esys_TR_SetAuth(esys_context,ESYS_TR_RH_OWNER,&authValue);

    TPM2B_PUBLIC *outPublic;
    TPM2B_CREATION_DATA *creationData;
    TPM2B_DIGEST *creationHash;
    TPMT_TK_CREATION *creationTicket;

    rc = Esys_CreatePrimary(
        esys_context,           // [in] esysContext
        ESYS_TR_RH_OWNER,       // [in] primaryHandle : hierarchy
        ESYS_TR_PASSWORD,       // [in] Session handle for authorization of primaryHandle
        ESYS_TR_NONE,           // Session handle 2
        ESYS_TR_NONE,           // Session handle 3
        &inSensitivePK,         // [in] inSensitive => sensitive data
        &inPublic,              // [in] public Template
        &outInfo,               // [in] data that will be included in the creation data this object to provide permanent
        &pcrSelection,          // [in] PCR that will be used in the creation process
        &primaryHandle,         // [in] object handle
        &outPublic,             // [out] Public portion of the created object 
        &creationData,          // [out] contains the creation data
        &creationHash,          // [out] Digest of creation data
        &creationTicket         // [out] ticket used to validate the creation of the object
    );

    if(!ErrorHandling(rc))
    {
        fprintf(stdout,"Error during the creation of the primary key\n");
        exit(EXIT_FAILURE);
    } else
    {
        fprintf(stdout,"Creation of the Key successfull \n");
    }

    rc = Esys_TR_SetAuth(esys_context,primaryHandle,&authValuePK);

    primaryKey->keyHandle = primaryHandle;
    primaryKey->creationData = creationData;
    primaryKey->creationHash = creationHash;
    primaryKey->creationTicket = creationTicket;
    primaryKey->outPublic = outPublic;
}

UINT16 vTPM::GetHashSize(TPMI_ALG_HASH hash)
{
    UINT16 hashSize = 0;

    switch (hash)
    {
    case TPM2_ALG_SHA1:
        hashSize = 20;
        break;
    case TPM2_ALG_SHA256:
        hashSize = 32;
        break;
    default:
        break;
    }

    return hashSize;
}

void vTPM::Test(TPM2B_PUBLIC inPublic)
{
    ESYS_TR objectHandle;

    Esys_LoadExternal(
        esys_context,
        ESYS_TR_NONE,
        ESYS_TR_NONE,
        ESYS_TR_NONE,
        nullptr,
        &inPublic,
        ESYS_TR_RH_OWNER,
        &objectHandle);
}
