#include "stdio.h"
#include "stdlib.h"
#include "verifier.h"
#include <iostream>
#include <cstdint>
#include "math.h"
#include "tss2/tss2_common.h"
#include "tss2/tss2_tpm2_types.h"
#include "tss2/tss2-tcti-tabrmd.h"
#include <inttypes.h>
#include <string.h>
#include <errno.h>
#include <iterator>
#include "MapKeyClass.h"
#include "tss2/tss2_mu.h"
#include <inttypes.h>
#include "tss2/tss2_esys.h"
#include <fstream>
#include <cstring>

using namespace std;

// Constructor
Verifier::Verifier(UINT32 id)
{
    // Initialization of the contexts
    Init_Tcti_Tabrmd_Context();
    Init_Esys_Context();
	identifier = id;
}

// Destructor
Verifier::~Verifier()
{
	// Finalization of the contexts
    Finalize_Esys_Context();
    Finalize_Tcti_Tabrmd_Context();
}

/*/////////////////////////////////////////////////////////
                    PRIVATE METHODS
/////////////////////////////////////////////////////////*/

void Verifier::Init_Tcti_Tabrmd_Context()
{
    TSS2_RC rc;
    size_t context_size;

    //Define allocation for the mssim TCTI
    /* 
        TSS2_RC Tss2_Tcti_Tabrmd_Init (TSS2_TCTI_CONTEXT *tcti_context, size_t *size, const char *conf)
        minimum size of the context => pass NULL to conf
        *tcti_context => desired connection properties for the TCTI
    */

    rc = Tss2_Tcti_Tabrmd_Init(nullptr, &context_size, NULL);
    if(!ErrorHandling(rc))
    {
        fprintf(stdout,"Failed to get allocation size for mssim TCTI \n");
        exit(EXIT_FAILURE);
    }
    
    tcti_context = (TSS2_TCTI_CONTEXT *) calloc(1,context_size);
    if(!ErrorHandling(rc))
    {
        fprintf(stdout, "Allocation for TCTI context failed: %s0 \n", strerror(errno));
        exit(EXIT_FAILURE);
    }

    rc = Tss2_Tcti_Tabrmd_Init(tcti_context,&context_size, NULL);
    if(!ErrorHandling(rc))
    {
        fprintf(stdout, "Failed to initialize tabrmd TCTI context: 0x%" PRIx32 "0\n", rc);
        free(tcti_context);
        exit(EXIT_FAILURE);
    }

    fprintf(stdout,"Initialization of the TCTI context successfull \n");
}

void Verifier::Init_Esys_Context()
{
    TSS2_RC rc;
    TSS2_ABI_VERSION abi_version = TSS2_ABI_VERSION_CURRENT;

    //Initialization of the ESYS context
    /*
        TSS2_RC Esys_Initialize( ESYS_CONTEXT **esysContext, TSS2_TCTI_CONTEXT *tcti, TSS2_ABI_VERSION *abiVersion);
        esysContext => reference to a pointer to the opaque ESYS_CONTEXT blob. It must be not NULL
        tcti => pointer to the TCTI context
        abiVersion => pointer to the ABI version that the application requests
    */
    rc = Esys_Initialize(&esys_context,tcti_context,&abi_version);
    if (rc != TSS2_RC_SUCCESS)
    {
        fprintf(stderr,"Failed to initialize the ESYS context: 0x%" PRIx32 "0 \n", rc);
        free(tcti_context);
        exit(EXIT_FAILURE);
    }

    fprintf(stdout,"Initialization of the ESYS context successfull \n");
}

void Verifier::Finalize_Tcti_Tabrmd_Context()
{
    Esys_Finalize(&esys_context);
    free(tcti_context);
    
    fprintf(stdout,"Finalization of the Esys context \n");
}

void Verifier::Finalize_Esys_Context()
{
    Tss2_Tcti_Finalize(tcti_context);

    fprintf(stdout,"Finalization of the TCTI context \n");
}


IntegrityRequest Verifier::CreateIntegrityRequest()
{
	IntegrityRequest integrityRequest;
	//TPM2B_DATA nonce 

	integrityRequest.nonce = {
		.size = 10,
		.buffer = {0,1,2,3,4,5,6,7,8,9}
	};

	cout << "Nonce = ";
	for (size_t i = 0; i < integrityRequest.nonce.size; i++)
    {
        fprintf(stdout, "%u", integrityRequest.nonce.buffer[i] );
    }
    cout << endl;

	integrityRequest.pcrSelection = {
        .count = 3,
        .pcrSelections = {
            {
                .hash = TPM2_ALG_SHA1,
                .sizeofSelect = 3,
                .pcrSelect = {1,0,0}
            },
            {
                .hash = TPM2_ALG_SHA1,
                .sizeofSelect = 3,
                .pcrSelect = {2,0,0}
            },
            {
                .hash = TPM2_ALG_SHA1,
                .sizeofSelect = 3,
                .pcrSelect = {4,0,0}
            }
        }
    };

    cout << "PCR selected: " << endl;
    UINT32 pcrNumber;
	for (int i = 0; i < integrityRequest.pcrSelection.count; i++)
	{	        
	  	for (int j = 0; j <integrityRequest.pcrSelection.pcrSelections[i].sizeofSelect; j++)
	    {
	        if(integrityRequest.pcrSelection.pcrSelections[i].pcrSelect[j]!=0)
	        {
	                // Extrapolation from the bitmap
	            pcrNumber = (j*8)+log2(integrityRequest.pcrSelection.pcrSelections[i].pcrSelect[j]);
	            fprintf(stdout,"PCR number #%u\n",pcrNumber);
	        }
	    }
	}
    return integrityRequest;
}

int Verifier::ErrorHandling(TSS2_RC rc)
{
    switch (rc)
    {
    case TSS2_RC_SUCCESS :
        return 1;
    case TSS2_ESYS_RC_BAD_REFERENCE:
        fprintf(stderr,"Bad Reference: the esysContext or required input pointers or required output handle references are NULL(0x%" PRIx32 " )\n",rc);
        return 0;
    case TSS2_ESYS_RC_MEMORY:
        fprintf(stderr,"he ESAPI cannot allocate enough memory for internal operations or return parameters(0x%" PRIx32 " )\n",rc);
        return 0;
    case TSS2_ESYS_RC_BAD_SEQUENCE:
        fprintf(stderr,"the context has an asynchronous operation already pending(0x%" PRIx32 " )\n",rc);
        return 0;
    case TSS2_ESYS_RC_INSUFFICIENT_RESPONSE:
        fprintf(stderr,"the TPM's response does not at least contain the tag, response length, and response code(0x%" PRIx32 " )\n",rc);
        return 0;
    case TSS2_ESYS_RC_MALFORMED_RESPONSE:
        fprintf(stderr,"the TPM's response is corrupted(0x%" PRIx32 " )\n",rc);
        return 0;
    case TSS2_ESYS_RC_RSP_AUTH_FAILED:
        fprintf(stderr,"the response HMAC from the TPM did not verify(0x%" PRIx32 " )\n",rc);
        return 0;
    case TSS2_ESYS_RC_MULTIPLE_DECRYPT_SESSIONS:
        fprintf(stderr,"more than one session has the 'decrypt' attribute bit set(0x%" PRIx32 " )\n",rc);
        return 0;
    case TSS2_ESYS_RC_MULTIPLE_ENCRYPT_SESSIONS:
        fprintf(stderr,"more than one session has the 'decrypt' attribute bit set(0x%" PRIx32 " )\n",rc);
        return 0;
    case TSS2_ESYS_RC_BAD_TR:
        fprintf(stderr,"ny of the ESYS_TR objects are unknown to the ESYS_CONTEXT or are of the wrong type or if required ESYS_TR objects are ESYS_TR_NONE(0x%" PRIx32 " )\n",rc);
        return 0;
    case TSS2_ESYS_RC_NO_ENCRYPT_PARAM:
        fprintf(stderr,"one of the sessions has the 'encrypt' attribute set and the command does not support encryption of the first response parameter(0x%" PRIx32 " )\n",rc);
        return 0;
    case TPM2_RC_NV_UNINITIALIZED:
        fprintf(stdout,"Read of NV location before it is written: 0x%" PRIx32 "\n", rc);
        return 0;
    case TSS2_RESMGR_RC_LAYER | TSS2_RESMGR_TPM_RC_LAYER:
        fprintf(stderr,"Command TPM2_GetCommandAuditDigest not supported by TPM : 0x%" PRIx32 " \n",rc);
        return 0;
    default:
        fprintf(stderr,"Error number : 0x%" PRIx32 " \n",rc);
        return 0;
    }
}

void Verifier::VerifyIntegrityReport(IntegrityReport integrityReport,Key rootKey, Key aik)
{
	// Return code
    TSS2_RC rc;
    ESYS_TR signKeyHandle;

   	TPM2B_ATTEST attestStructure = integrityReport.attestStructure_VM;
	TPMT_SIGNATURE signatureStructure = integrityReport.signatureStructure_VM;
	TPML_DIGEST pcrValues = integrityReport.pcrValues_VM;

    rc = Esys_LoadExternal(
        esys_context,
        ESYS_TR_NONE,
        ESYS_TR_NONE,
        ESYS_TR_NONE,
        NULL,
        aik.outPublic,
        TPM2_RH_NULL,
        &signKeyHandle
    );

    if(!ErrorHandling(rc))
    {
        fprintf(stdout,"Error during the load operation of the Public Key\n");
        exit(EXIT_FAILURE);
    } else
    {
        fprintf(stdout,"Verifer loaded public key \n");
    }

    TPM2B_AUTH authKey = {
        .size = 5,
        .buffer = {1,2,3,4,5}
    };

    rc = Esys_TR_SetAuth(esys_context,signKeyHandle, &authKey);

    // Now that we have created the structure we need to calculate the hash in order to verify the signature
    TPMI_ALG_HASH hashAlgorithm = aik.outPublic->publicArea.nameAlg;
    TPM2B_DIGEST* quoteHash;
    TPMT_TK_HASHCHECK* quoteHashValidation;
    TPM2B_MAX_BUFFER bufferToHash = {.size=attestStructure.size, .buffer={0}};
    size_t offset = 0;

    cout << "*** Veryifing VM Remote Atterstation ***" << endl;
    cout << "Received structure" << endl;
    for (size_t i = 0; i < attestStructure.size; i++)
    {
        fprintf(stdout, "%" PRIX8 "", attestStructure.attestationData[i] );
    }
    cout << endl;

    // Copying the attestStructure in a suitable structure to be hashed
    for (int i = 0; i < attestStructure.size; ++i)
    {
        bufferToHash.buffer[i]=attestStructure.attestationData[i];
    }

    cout<<"bufferToHash"<<endl;
    for (int i = 0; i < bufferToHash.size ; i++)
    {
        fprintf(stdout, "%" PRIX8 "", bufferToHash.buffer[i] );
    }
    cout << endl;

    rc = Esys_Hash(
        esys_context,
        ESYS_TR_NONE,
        ESYS_TR_NONE,
        ESYS_TR_NONE,
        &bufferToHash,
        hashAlgorithm, // to be taken from signing scheme
        TPM2_RH_NULL,
        &quoteHash,
        &quoteHashValidation
    );

    cout<<"\nHashed structure"<<endl;
    for (int i = 0; i < quoteHash->size ; i++)
    {
        fprintf(stdout, "%" PRIX8 "", quoteHash->buffer[i] );
    }
    cout << endl;

    fprintf(stdout,"Signature received\n");
    for (int i = 0; i < signatureStructure.signature.rsapss.sig.size ; i++)
    {
        fprintf(stdout,"%" PRIX8 " ",signatureStructure.signature.rsapss.sig.buffer[i]);
    }
    cout << endl;

    TPMT_TK_VERIFIED *validationSignature;
    rc = Esys_VerifySignature(
        esys_context,
        signKeyHandle,
        ESYS_TR_NONE,
        ESYS_TR_NONE,
        ESYS_TR_NONE,
        quoteHash,
        &signatureStructure,
        &validationSignature
    );

    if(!ErrorHandling(rc))
    {
        fprintf(stdout,"Error during signature verification\n");
    } else
    {
        fprintf(stdout,"Signature Verified \n");
    }

    // Now that the signature is verified we can verify if the VM can be considered as TRUST
    cout << "PCR values: " << endl;

    for (int i = 0; i < pcrValues.count; i++)
    {
    	cout << "PCR #" << i << " " ;
    	for (int j = 0; j < pcrValues.digests[i].size; j++)
    	{
    		fprintf(stdout,"%" PRIX8 " ",pcrValues.digests[i].buffer[j]);
    	}
    	cout << endl ;
    }

	attestStructure = integrityReport.attestStructure_VMM;
	signatureStructure = integrityReport.signatureStructure_VMM;
	pcrValues = integrityReport.pcrValues_VMM;
	bufferToHash = {.size=attestStructure.size, .buffer={0}};

	cout << "*** Veryifing VMM Remote Atterstation ***" << endl;
    cout << "Received structure" << endl;
    for (size_t i = 0; i < attestStructure.size; i++)
    {
        fprintf(stdout, "%" PRIX8 "", attestStructure.attestationData[i] );
    }
    cout << endl;

    // Copying the attestStructure in a suitable structure to be hashed
    for (int i = 0; i < attestStructure.size; ++i)
    {
        bufferToHash.buffer[i]=attestStructure.attestationData[i];
    }

    cout<<"bufferToHash"<<endl;
    for (int i = 0; i < bufferToHash.size ; i++)
    {
        fprintf(stdout, "%" PRIX8 "", bufferToHash.buffer[i] );
    }
    cout << endl;

    rc = Esys_Hash(
        esys_context,
        ESYS_TR_NONE,
        ESYS_TR_NONE,
        ESYS_TR_NONE,
        &bufferToHash,
        hashAlgorithm, // to be taken from signing scheme
        TPM2_RH_NULL,
        &quoteHash,
        &quoteHashValidation
    );

    cout<<"\nHashed structure"<<endl;
    for (int i = 0; i < quoteHash->size ; i++)
    {
        fprintf(stdout, "%" PRIX8 "", quoteHash->buffer[i] );
    }
    cout << endl;

    fprintf(stdout,"Signature received\n");
    for (int i = 0; i < signatureStructure.signature.rsapss.sig.size ; i++)
    {
        fprintf(stdout,"%" PRIX8 " ",signatureStructure.signature.rsapss.sig.buffer[i]);
    }
    cout << endl;

    rc = Esys_VerifySignature(
        esys_context,
        signKeyHandle,
        ESYS_TR_NONE,
        ESYS_TR_NONE,
        ESYS_TR_NONE,
        quoteHash,
        &signatureStructure,
        &validationSignature
    );

    if(!ErrorHandling(rc))
    {
        fprintf(stdout,"Error during signature verification\n");
    } else
    {
        fprintf(stdout,"Signature Verified \n");
    }

    // Now that the signature is verified we can verify if the VM can be considered as TRUST
    cout << "PCR values: " << endl;

    for (int i = 0; i < pcrValues.count; i++)
    {
    	cout << "PCR #" << i << " " ;
    	for (int j = 0; j < pcrValues.digests[i].size; j++)
    	{
    		fprintf(stdout,"%" PRIX8 " ",pcrValues.digests[i].buffer[j]);
    	}
    	cout << endl ;
    }

    // Now we have to reproduce all the logs received to check if the values are matching
    cout << "Checking the logs" << endl ;

    //In the file Hashedlogs there is the computed hash of the three files binaries, firmware and bootloader logs
    
    fstream hashfile;
    string line;
    hashfile.open("hashedlogs.log");
	while(!hashfile.eof())
    {
        getline(hashfile,line);
        cout << line << endl;
    }

    //PCRs now have the same expected value
}